import os
import tempfile
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple

import jinja2
import typer
from pydantic import ValidationError
from rich import print

from itmpl import config, global_vars, metadata, tree_utils, utils


class DuplicateTemplateError(Exception):
    """Exception raised when there are duplicate templates."""

    def __init__(
        self,
        duplicate_templates: Dict[str, Tuple[Path, str, List[str]]],
    ) -> None:
        super().__init__()
        self.duplicate_templates = duplicate_templates


class IgnoreUndefined(jinja2.Undefined):
    """Ignore undefined variables."""

    def __str__(self):
        if not self._undefined_name:
            return ""
        return "{{ " + self._undefined_name + " }}"


class TemplatingException(Exception):
    """Exception raised when there is an error with the templating."""


def get_templates_in_dir(directory: Path) -> Dict[str, Tuple[Path, str, List[str]]]:
    """Return a list of templates in a directory with their descriptions."""
    templates = {}
    for path in directory.iterdir():
        if not path.is_dir():
            continue

        meta = metadata.read_itmpl_toml(path / ".itmpl.toml")
        templates[path.name] = (
            path,
            meta.metadata.template_description,
            meta.metadata.template_requirements,
        )

    return templates


def get_template_options() -> Dict[str, Tuple[Path, str, List[str]]]:
    """Return a list of template options and their descriptions."""
    c = config.read_config()
    default_template_options = get_templates_in_dir(global_vars.TEMPLATES_DIR)
    extra_template_options = get_templates_in_dir(c.extra_templates_dir)

    # Find the intersection of the two sets of templates
    duplicate_template_keys = (
        default_template_options.keys() & extra_template_options.keys()
    )
    duplicate_templates = {
        k: v
        for k, v in default_template_options.items()
        if k in duplicate_template_keys
    }

    if duplicate_templates:
        raise DuplicateTemplateError(duplicate_templates)

    return {**default_template_options, **extra_template_options}


def _setup_itmpl_module(directory: Path) -> Optional[ModuleType]:
    """Import the .itmpl.py file in the template directory."""
    itmpl_file = directory / ".itmpl.py"

    if not itmpl_file.exists():
        return None

    try:
        return utils.import_external_module(itmpl_file)
    except Exception as e:
        raise TemplatingException(f"Error when importing .itmpl.py: {e}") from e


def get_default_variables(project_name: str) -> Dict[str, Any]:
    return {
        "project_name": project_name,
        "project_title": project_name.replace("-", " ").replace("_", " ").title(),
    }


def get_toml_variables(temp_directory: Path) -> Dict[str, Any]:
    """Get extra variables from the .itmpl.toml file in the template directory."""
    try:
        meta = metadata.read_itmpl_toml(temp_directory / ".itmpl.toml")
    except ValidationError as e:
        raise TemplatingException(f"Error when validating .itmpl.toml: {e}") from e
    except Exception as e:
        raise TemplatingException(f"Error when reading .itmpl.toml: {e}") from e

    return meta.variables


def get_python_variables(
    temp_directory: Path,
    project_name: str,
    destination: Path,
    variables: Dict[str, Any],
) -> Dict[str, str]:
    """Get extra variables from the .itmpl.py file in the template directory."""
    module = _setup_itmpl_module(temp_directory)

    if not module or not hasattr(module, "get_variables"):
        return {}

    try:
        return module.get_variables(
            project_name,
            destination,
            variables,
        )
    except Exception as e:
        raise TemplatingException(
            f"Error when getting extra variables from .itmpl.py: {e}"
        ) from e


def run_post_script(
    project_name: str,
    final_directory: Path,
    variables: Dict[str, str],
) -> Dict[str, str]:
    """Run the post script in the .itmpl.py file in the template directory."""
    module = _setup_itmpl_module(final_directory)

    if not module or not hasattr(module, "post_script"):
        return {}

    try:
        return module.post_script(
            project_name,
            final_directory,
            variables,
        )
    except Exception as e:
        raise TemplatingException(
            f"Error when running post script from .itmpl.py: {e}"
        ) from e


def template_directory(
    dir_path: Path,
    variables: Dict[str, Any],
    ignore_undefined: bool = False,
) -> None:
    """Template the contents of a directory using Jinja. Both file contents and
    filenames are templated."""
    directories_to_rename = []

    for root, dirs, files in os.walk(dir_path):
        root = Path(root)

        for directory in dirs:
            directories_to_rename.append(root / directory)

        for file in files:
            file_path = root / file

            # Ignore .itmpl files
            if file_path.name.startswith(".itmpl"):
                continue

            # Template the file's contents
            try:
                contents_template = jinja2.Template(
                    Path(file_path).read_text(),
                    undefined=(
                        IgnoreUndefined if ignore_undefined else jinja2.StrictUndefined
                    ),
                    keep_trailing_newline=True,
                )
            except UnicodeDecodeError:
                # Not a unicode file, so skip it
                continue
            rendered = contents_template.render(**variables)
            Path(file_path).write_text(rendered)

            # Rename the file
            filename_template = jinja2.Template(file)
            rendered = filename_template.render(**variables)
            file_path.rename(root / rendered)

    # Rename directories
    # Note: we have to reverse the list of directories to rename because
    # otherwise we might rename a parent directory, and then try to rename its
    # children using an incorrect path.
    for directory in reversed(directories_to_rename):
        root = directory.parent
        dirname_template = jinja2.Template(directory.name)
        rendered = dirname_template.render(**variables)
        directory.rename(root / rendered)


def render_template(
    project_name: str,
    template: str,
    destination: Path,
    template_path: Path,
    prompt_if_duplicates: bool = True,
):
    default_variables = {
        **get_default_variables(project_name=project_name),
        **global_vars.VARIABLES,
    }

    with tempfile.TemporaryDirectory() as tempdir:
        temp_project_dir = Path(tempdir) / template
        tree_utils.copy_tree(
            template_path,
            temp_project_dir,
        )

        toml_variables = get_toml_variables(temp_project_dir)
        python_variables = get_python_variables(
            temp_directory=temp_project_dir,
            project_name=project_name,
            destination=destination,
            variables={**default_variables, **toml_variables},
        )

        variables = {**default_variables, **toml_variables, **python_variables}
        template_directory(temp_project_dir, variables, ignore_undefined=True)

        duplicates = list(
            tree_utils.find_duplicates(
                temp_project_dir,
                destination,
            ),
        )

        if duplicates and prompt_if_duplicates:
            print(
                "[red]The following files already exist and will be overwritten:[/red]",
            )
            print("\n".join([str(d.resolve()) for d in duplicates]))

            typer.confirm("Continue?", abort=True)

        tree_utils.copy_tree(temp_project_dir, destination)
        new_variables = run_post_script(
            project_name=project_name,
            final_directory=destination,
            variables=variables.copy(),
        )

        # If the post script returns new variables, template the directory again with
        # the new variables. This time, we don't ignore undefined variables, so that
        # any extraneous Jinja is ignored.
        if new_variables:
            try:
                template_directory(destination, new_variables, ignore_undefined=False)
            except jinja2.exceptions.UndefinedError as e:
                raise TemplatingException(
                    f"Error when templating directory: {e}"
                ) from e

        tree_utils.recursive_delete(destination, ".itmpl*")
        tree_utils.recursive_delete(destination, "__pycache__")



import importlib.util
import subprocess
import sys
from pathlib import Path
from types import ModuleType
from typing import Iterable, List, Tuple

from rich.table import Table


def import_external_module(module_path: Path) -> ModuleType:
    """Import a module from an external path."""
    spec = importlib.util.spec_from_file_location("itmpl", module_path)

    if spec is None:
        raise ValueError(f"No module found at path: {module_path}")

    module = importlib.util.module_from_spec(spec)

    if spec.loader is None:
        raise ValueError(f"No loader found for module: {module_path}")

    try:
        spec.loader.exec_module(module)
    except FileNotFoundError:
        raise ValueError(f"No module found at path: {module_path}")

    return module


def construct_table_from_templates(
    templates: Iterable[Tuple[Path, str, List[str]]],
) -> Table:
    """Construct a Rich table from a list of templates."""
    table = Table(show_header=True, header_style="bold")
    table.add_column("Template", justify="left", no_wrap=True, header_style="blue")
    table.add_column("Description")
    table.add_column("Requirements")

    for template in templates:
        table.add_row(template[0].name, template[1], ", ".join(template[2]))

    return table


def install_dependencies(dependencies: List[str]):
    """Install dependencies."""
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", *dependencies],
    )



import json
import subprocess
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional

import typer
import yaml


def get_variables(
    project_name: str,
    destination: Path,
    variables: Dict[str, Any],
) -> Dict[str, Any]:
    project_description = typer.prompt("Project description")
    project_author = typer.prompt("Project author")
    project_license = typer.prompt("Project license", default="MIT")
    python_version = typer.prompt("Python version", default="3.11")
    return {
        "project_description": project_description,
        "project_author": project_author,
        "project_license": project_license,
        "python_version": python_version,
    }


def _add_dev_dependencies(final_directory: Path) -> None:
    dev_dependencies = final_directory / ".itmpl.requirements.dev.txt"

    if not dev_dependencies.exists():
        return

    print("Adding dev dependencies...", end=" ", flush=True)
    subprocess.run(
        [
            "poetry",
            "add",
            "-G",
            "dev",
            *dev_dependencies.read_text().splitlines(),
        ],
        cwd=final_directory,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    print("done!")


def _add_dependencies(final_directory: Path) -> None:
    dependencies = final_directory / ".itmpl.requirements.txt"

    if not dependencies.exists():
        return

    print("Adding dependencies...", end=" ", flush=True)
    subprocess.run(
        [
            "poetry",
            "add",
            *dependencies.read_text().splitlines(),
        ],
        cwd=final_directory,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    print("done!")


def _get_package_versions(package_names: List[str]) -> Dict[str, str]:
    """Get installed package versions."""
    installed_packages_json = subprocess.run(
        ["poetry", "run", "pip", "list", "--format=json"],
        capture_output=True,
        text=True,
        check=True,
    ).stdout

    installed_packages = json.loads(installed_packages_json)

    return {
        package["name"]: package["version"]
        for package in installed_packages
        if package["name"] in package_names
    }


def _get_pre_commit_config_package_versions(directory: Path) -> Dict[str, str]:
    """Get package versions from the pre-commit config file."""
    pre_commit_config = directory / ".pre-commit-config.yaml"

    if not pre_commit_config.exists():
        return {}

    with pre_commit_config.open() as f:
        config = yaml.safe_load(f)

    package_names = []

    for repo in config["repos"]:
        for hook in repo["hooks"]:
            package_names.append(hook["name"])

    return _get_package_versions(package_names)


def post_script(
    project_name: str,
    final_directory: Path,
    variables: Dict[str, Any],
) -> Optional[Dict[str, Any]]:
    # Deactivate virtual environment
    try:
        subprocess.run(["deactivate"], check=False)
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    # Check if poetry is installed
    try:
        subprocess.run(["poetry", "--version"], check=True, stdout=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        # Install poetry
        print("Poetry is not installed, installing Poetry...", end=" ", flush=True)
        subprocess.run(
            [sys.executable, "-m", "pip", "install", "poetry"],
            check=True,
            stdout=subprocess.DEVNULL,
        )
        print("done!")

    # Add dependencies
    _add_dependencies(final_directory)
    _add_dev_dependencies(final_directory)

    # Install dependencies
    print("Installing dependencies...", end=" ", flush=True)
    subprocess.run(
        ["poetry", "install"],
        cwd=final_directory,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    print("done!")

    # Get package versions of packages used in pre-commit config
    package_versions = _get_pre_commit_config_package_versions(final_directory)

    # Add package versions to variables
    version_template_strings = {
        f"{package_name}_version": version
        for package_name, version in package_versions.items()
    }
    variables.update(version_template_strings)

    return variables



import shutil
from pathlib import Path
from typing import Callable, Optional


def find_duplicates(
    source: Path,
    destination: Path,
    ignore: Optional[Callable[[Path], bool]] = None,
):
    """Find all the files in the destination directory that are present in the
    source directory. Unlike shutil.copytree, this function will not overwrite
    existing files and will create the destination directory if it does not
    exist.
    """
    ignore = ignore or (lambda p: False)

    for item in source.iterdir():
        if ignore(item):
            # Item is ignored
            continue
        elif not (destination / item.name).exists():
            # Item does not exist in destination
            continue
        elif item.is_dir():
            # Item is a directory, so recurse
            yield from find_duplicates(item, destination / item.name)
        elif (destination / item.name).exists():
            # Item is a file and exists in destination
            yield destination / item.name


def copy_tree(
    source: Path,
    destination: Path,
    ignore: Optional[Callable[[Path], bool]] = None,
):
    """Copy a tree of files from source to destination. Unlike shutil.copytree,
    this function will not overwrite existing files and will create the
    destination directory if it does not exist.
    """
    ignore = ignore or (lambda p: False)

    for item in source.iterdir():
        if ignore(item):
            continue
        elif item.is_dir():
            copy_tree(item, destination / item.name)
        else:
            destination.mkdir(parents=True, exist_ok=True)
            shutil.copy2(item, destination / item.name)


def recursive_delete(directory: Path, glob: str) -> None:
    """Delete all files in a directory matching a glob."""
    for file in directory.rglob(glob):
        if file.is_file():
            file.unlink()
        else:
            shutil.rmtree(file)



import enum
from pathlib import Path

import typer
from pydantic import BaseModel
from pydantic.fields import ModelField
from rich import print
from typer import Typer

from itmpl.global_vars import APP_DIR

app = Typer()


CONFIG_PATH = APP_DIR / "config.json"


class Config(BaseModel):
    """Configuration for iTmpl."""

    extra_templates_dir: Path = APP_DIR / "templates"


ConfigOption = enum.Enum("ConfigOption", {k: k for k in Config.__fields__})


def read_config() -> Config:
    if not CONFIG_PATH.exists():
        return Config()
    return Config.parse_file(CONFIG_PATH)


def write_config(config: Config):
    with CONFIG_PATH.open("w") as f:
        f.write(config.json())


@app.command("show")
def show(option: ConfigOption = typer.Argument(None)):  # type: ignore
    """Show the current configuration."""
    config = read_config()
    if option:
        print(config.__getattribute__(option.value))
        return

    print(config.json(indent=4))


@app.command("set")
def set_(
    option: ConfigOption,  # type: ignore
    value: str,
):
    """Set a configuration option."""
    config = read_config()

    # Some custom validation using Pydantic's ModelField validation
    model_field: ModelField = config.__fields__[option.value]
    other_attrs = {k: v for k, v in config.__dict__.items() if k != option.value}
    new, error = model_field.validate(value, other_attrs, loc=option.value, cls=Config)

    if error:
        raise typer.BadParameter(
            f"{value} is not a valid {model_field.type_.__name__}",
            param_hint=option.value,
        )

    if new and model_field.type_ == Path and not (new.exists() and new.is_dir()):
        raise typer.BadParameter(
            f"{value} is not a valid directory",
            param_hint=option.value,
        )

    config.__setattr__(option.value, new)
    write_config(config)

    print(f"Set [green]{option.value}[/green] to [green]{new}[/green]")


@app.command()
def get(option: ConfigOption):  # type: ignore
    """Get a configuration option."""
    config = read_config()
    print(
        f"[green]{option.value}[/green] is set to "
        f"[green]{config.__getattribute__(option.value)}[/green]"
    )


@app.command()
def reset():
    """Reset the configuration to the default values."""
    typer.confirm("Are you sure you want to reset the configuration?", abort=True)
    write_config(Config())
    print("Reset configuration to default values.")



import subprocess
from pathlib import Path
from typing import Optional

import typer
from rich import print
from typer import Typer

from itmpl import config, global_vars, templating, utils

app = Typer()
app.add_typer(config.app, name="config", help="Manage iTmpl configuration.")


@app.command("list")
def list_():
    """List all available templates."""
    print(
        utils.construct_table_from_templates(
            templating.get_template_options().values(),
        ),
    )


@app.command()
def new(
    template: str,
    name: str,
    path: Path = typer.Option(
        Path("."),
        "--path",
        "-p",
        exists=True,
        help="The path to create the project in.",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Overwrite any files that already exist without prompting.",
    ),
):
    """Create a new project from a template.

    Parameters
    ----------
    template : str
        The name of the template to use.
    name : str
        The name of the project to create.
    path : Path
        The path to create the project in.
    force : bool
        If True, overwrite any files that already exist without prompting.
    """
    try:
        template_options = templating.get_template_options()
    except templating.DuplicateTemplateError as e:
        print("[red]Duplicate templates found:[/red]")
        print(utils.construct_table_from_templates(e.duplicate_templates.values()))
        print("[red]Please remove the duplicates and try again.[/red]")
        raise typer.Exit(1)

    if template not in template_options:
        print(
            f"[red]Template [white]{template}[/white] not found. "
            f"Available templates:[/red]"
        )
        print(utils.construct_table_from_templates(template_options.values()))
        raise typer.Exit(1)

    # Allow the user to template in this directory
    path = path.resolve()
    if path.name == name:
        path = path.parent

    destination = path / name

    template_path = template_options[template][0]

    try:
        templating.render_template(
            project_name=name,
            template=template,
            destination=destination,
            template_path=template_path,
            prompt_if_duplicates=not force,
        )
    except templating.TemplatingException as e:
        print(f"[red]Error when templating project:[/red] {e}")
        raise typer.Exit(1)

    print(f"Created [green]{template}[/green] project at [green]{destination}[/green]")


@app.command()
def deps(template: Optional[str] = typer.Argument(None)):
    """Install dependencies for the specified template. If no template is specified,
    install dependencies for all templates."""
    try:
        template_options = templating.get_template_options()
    except templating.DuplicateTemplateError as e:
        print("[red]Duplicate templates found:[/red]")
        print(utils.construct_table_from_templates(e.duplicate_templates.values()))
        print("[red]Please remove the duplicates and try again.[/red]")
        raise typer.Exit(1)

    if template is not None and template not in template_options:
        print(
            f"[red]Template [white]{template}[/white] not found. "
            f"Available templates:[/red]"
        )
        print(utils.construct_table_from_templates(template_options.values()))
        raise typer.Exit(1)

    if template is None:
        templates_with_dependencies = [
            (template_name, dependencies)
            for template_name, (_, _, dependencies) in template_options.items()
        ]
    else:
        templates_with_dependencies = [(template, template_options[template][2])]

    to_install = [
        (template_name, dependencies)
        for template_name, dependencies in templates_with_dependencies
        if dependencies
    ]

    if not to_install:
        print("[green]No dependencies to install.[/green]")
        raise typer.Exit(0)

    for template_name, dependencies in to_install:
        print(
            f"[green]Installing dependencies for template "
            f"[white]{template_name}[/white][/green]"
        )
        try:
            utils.install_dependencies(dependencies)
        except subprocess.CalledProcessError as e:
            print(
                f"[red]Error installing dependencies for template "
                f"[white]{template_name}[/white]: {e}[/red]"
            )

    print("[green]Done.[/green]")


@app.callback()
def create_directories():
    """Create directories used by iTmpl. This is called automatically when iTmpl is
    run."""
    global_vars.APP_DIR.mkdir(parents=True, exist_ok=True)
    global_vars.TEMPLATES_DIR.mkdir(parents=True, exist_ok=True)
    c = config.read_config()
    c.extra_templates_dir.mkdir(parents=True, exist_ok=True)


if __name__ == "__main__":
    app()



import subprocess
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import typer


class DependencyManager(Enum):
    POETRY = "poetry"
    PIP = "pip"
    NONE = "none"


def get_variables(
    project_name: str,
    destination: Path,
    variables: Dict[str, Any],
) -> Dict[str, Any]:
    author_name = typer.prompt("Author name")
    site_url = typer.prompt("Site URL")

    if site_url.endswith("/"):
        site_url = site_url[:-1]

    light_mode_primary_colour = typer.prompt(
        "Light mode primary colour",
        default="#2e69dc",
    ).replace("#", "")
    dark_mode_primary_colour = typer.prompt(
        "Dark mode primary colour",
        default="#19428e",
    ).replace("#", "")
    return {
        "author_name": author_name,
        "site_url": site_url,
        "light_mode_primary_colour": light_mode_primary_colour,
        "dark_mode_primary_colour": dark_mode_primary_colour,
    }


def _add_dependencies_poetry(
    dependencies: List[str],
    final_directory: Path,
) -> None:
    print("Adding Poetry dev dependencies...", end=" ", flush=True)
    subprocess.run(
        [
            "poetry",
            "add",
            "-G",
            "dev",
            *dependencies,
        ],
        cwd=final_directory,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    print("done!")

    # Install dependencies
    print("Installing Poetry dev dependencies...", end=" ", flush=True)
    subprocess.run(
        ["poetry", "install"],
        cwd=final_directory,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    print("done!")


def _add_dependencies_requirements(
    dependencies: List[str],
    final_directory: Path,
) -> None:
    print("Adding dependencies...", end=" ", flush=True)
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            *dependencies,
        ],
        cwd=final_directory,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    print("done!")

    docs_requirements_file = final_directory / "requirements.docs.txt"
    dev_requirements_file = final_directory / "requirements.dev.txt"
    requirements_file = final_directory / "requirements.txt"

    if docs_requirements_file.exists():
        print(
            "Adding dev dependencies to requirements.docs.txt...",
            end=" ",
            flush=True,
        )
        with dev_requirements_file.open("a") as f:
            f.write("\n".join(dependencies))
        print("done!")
    elif dev_requirements_file.exists():
        print("Adding dev dependencies to requirements.dev.txt...", end=" ", flush=True)
        with dev_requirements_file.open("a") as f:
            f.write("\n".join(dependencies))
        print("done!")
    else:
        if not requirements_file.exists():
            requirements_file.touch()

        print("Adding dev dependencies to requirements.txt...", end=" ", flush=True)
        with requirements_file.open("a") as f:
            f.write("\n".join(dependencies))
        print("done!")


def post_script(
    project_name: str,
    final_directory: Path,
    variables: Dict[str, Any],
) -> Optional[Dict[str, Any]]:
    # Get pip path
    dependency_method = typer.prompt(
        "How would you like to add dependencies? [pip, poetry, none]",
        type=DependencyManager,
    )

    dependencies = (
        (final_directory / ".itmpl.requirements.dev.txt").read_text().splitlines()
    )

    if dependency_method == DependencyManager.POETRY:
        _add_dependencies_poetry(dependencies, final_directory)
    elif dependency_method == DependencyManager.PIP:
        _add_dependencies_requirements(dependencies, final_directory)
    elif dependency_method == DependencyManager.NONE:
        print(
            "Skipping dependency installation...\n"
            "You will need to install the following dependencies manually:"
        )
        print("\n".join(dependencies))
    else:
        raise ValueError(f"Unknown dependency method: {dependency_method}")

    return {}



from datetime import datetime
from pathlib import Path
from typing import Dict

import typer

APP_DIR: Path = Path(typer.get_app_dir("itmpl"))

TEMPLATES_DIR: Path = Path(__file__).parent / "templates"

VARIABLES: Dict[str, str] = {
    "current_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    "current_year": datetime.now().strftime("%Y"),
}



from pathlib import Path
from typing import Any, Dict, List, Optional

try:
    import tomli
except ImportError:
    # Python 3.11
    import tomllib as tomli  # type: ignore

from pydantic import BaseModel


class ItmplMetadata(BaseModel):
    """Metadata from the .itmpl.toml file."""

    template_description: Optional[str] = None
    template_requirements: Optional[List[str]] = []


class ItmplToml(BaseModel):
    """The whole contents of the .itmpl.toml file."""

    metadata: ItmplMetadata = ItmplMetadata()
    variables: Dict[str, Any] = {}


def read_itmpl_toml(path: Path) -> ItmplToml:
    if not path.exists():
        return ItmplToml()
    return ItmplToml.parse_obj(tomli.loads(path.read_text(encoding="utf-8")))
